@@ -24,7 +24,7 @@ static const unsigned char usb_kbd_keycode[256] = {
 
 static const unsigned char kbd_lower[127] = {
 '\0', '\0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', // 13
-'\0', '\t', 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', // 27
+'\x8', '\t', 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', // 27
 '\n', '\0', 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\'', '`', // 41
 '\0', '\\', 'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/', '\0', '*', // 55
 '\0', ' ', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0',
@@ -34,7 +34,7 @@ static const unsigned char kbd_lower[127] = {
 };
 static const unsigned char kbd_upper[127] = {
 '\0', '\0', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '_', '+',
-'\0', '\0', 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '{', '}',
+'\x8', '\0', 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '{', '}',
 '\0', '\0', 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', ':', '"', '~',
 '\0', '|', 'Z', 'X', 'C', 'V', 'B', 'N', 'M', '<', '>', '?', '\0', '\0',
 '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0',
@@ -56,26 +56,32 @@ char code_to_key(bool shifted, unsigned int kcode) {
 }
 
 void cleanup(usb_dev_handle *handle) {
-  int ret = usb_release_interface(handle, 0);
+  usb_release_interface(handle, 0);
   usb_close(handle);
 }
 
-usb_dev_handle* _find_device (int vendor, int product, int iface) {
+usb_dev_handle* _find_device (int vendor, int product, int busnum, int devnum, int iface) {
   struct usb_bus *bus;
   struct usb_device *device;
   usb_dev_handle *handle;
   int ret;
+  static bool initialized = 0;
 
-  usb_init();
+  if(!initialized) {
+    usb_init();          // do once per process instantiation
+    initialized = 1;
+  }
   usb_find_busses();
   usb_find_devices();
   // TODO have my own globals for init?
   bus = usb_get_busses();
 
   for(; bus; bus = bus->next) {
+    if (busnum >= 0 && atoi(bus->dirname) != busnum) continue;
     for(device = bus->devices; device; device = device->next) {
-      if(device->descriptor.idVendor == vendor &&
-          device->descriptor.idProduct == product) {
+      if (devnum >= 0 && device->devnum != devnum) continue;
+      if ( (vendor  < 0 || device->descriptor.idVendor  == vendor)  &&
+           (product < 0 || device->descriptor.idProduct == product) ) {
         handle = usb_open(device);
         // XXX non-portable, and I guess we don't need to retry
         usb_detach_kernel_driver_np(handle, iface);
@@ -88,19 +94,34 @@ usb_dev_handle* _find_device (int vendor, int product, int iface) {
       }
     }
   }
-  croak("failed to find device %x %x", vendor, product);
+
+  char vStr[30] = "", pStr[30] = "", bStr[30] = "", dStr[30] = "";
+  if (vendor  >= 0) snprintf(vStr, sizeof(vStr), " vendor=0x%x",  vendor);
+  if (product >= 0) snprintf(pStr, sizeof(pStr), " product=0x%x", product);
+  if (busnum  >= 0) snprintf(bStr, sizeof(bStr), " busnum=%d",    busnum);
+  if (devnum  >= 0) snprintf(dStr, sizeof(dStr), " devnum=%d",    devnum);
+  croak("failed to find any device matching:%s%s%s%s", vStr, pStr, bStr, dStr);
 }
 
-SV*	create (char* class, int vendor_id, int product_id, int iface) {
-	SV*        obj_ref = newSViv(0);
-	SV*        obj = newSVrv(obj_ref, class); // creates the blessed reference
+int fetchInt(HV* hash, const char* key, int len, int defaultVal) {
+  SV** valp = hv_fetch(hash, key, len, 0);
+  return valp ? (int)SvIV(*valp) : defaultVal;
+}
 
-  struct usb_dev_handle *handle = _find_device(vendor_id, product_id, iface);
+void _usb_init (SV* obj) {
+  SV** selectorp = hv_fetchs((HV*)SvRV(obj), "selector", 0);
+  if (!selectorp) croak("sanity failure: no selector in $self");
+  HV* selector = (HV*)SvRV(*selectorp);
+
+  int vendor  = fetchInt(selector, "vendor",  6, -1);
+  int product = fetchInt(selector, "product", 7, -1);
+  int busnum  = fetchInt(selector, "busnum",  6, -1);
+  int devnum  = fetchInt(selector, "devnum",  6, -1);
+  int iface   = fetchInt(selector, "iface",   5,  0);
+
+  struct usb_dev_handle *handle = _find_device(vendor, product, busnum, devnum, iface);
   // fprintf(stderr, "got handle %d\n", handle);
-
-	sv_setiv(obj, (IV) handle);
-	SvREADONLY_on(obj);
-	return(obj_ref);
+  hv_store((HV*)SvRV(obj), "handle", 6, newSViv((IV)handle), 0);
 }
 
 #define PACKET_LEN 8
@@ -112,21 +133,26 @@ void _dump_packet(const char* packet) {
   fprintf(stderr, "\n");
 }
 
-// XXX hmm, this scheme can't detect a second key while you're holding
-// one down.  I guess "don't do that" applies.
 void _keycode(SV* obj, int timeout) {
-	Inline_Stack_Vars;
-  usb_dev_handle* handle = (usb_dev_handle*) SvIV(SvRV(obj));
+  Inline_Stack_Vars;
+  int prevKeydown = fetchInt((HV*)SvRV(obj), "prevKeydown", 11, 0);
+  usb_dev_handle* handle = (usb_dev_handle*) fetchInt((HV*)SvRV(obj), "handle", 6, 0);
 
-	Inline_Stack_Reset;
+  Inline_Stack_Reset;
 
   // XXX right_super is a lot bigger than 255 for some reason?
   unsigned char packet[PACKET_LEN];
   // croak("handle is %d", handle);
-  int ret = usb_interrupt_read(handle, 0x81, packet, PACKET_LEN, timeout);
-  if(ret > 0 && ! packet[3]) {
+  int latest = usb_interrupt_read(handle, 0x81, packet, PACKET_LEN, timeout);
+
+  // find the most recent keydown (there could be several)
+  for (--latest; latest > 1 && packet[latest] == 0; --latest) {}
+
+  // if keydown found and is not a dupe (dupes happen with simultaneous keydowns)
+  if(latest > 1 && packet[latest] != prevKeydown) {
     // fprintf(stderr, "read %d bytes\n", ret);
-    Inline_Stack_Push(sv_2mortal(newSViv(usb_kbd_keycode[packet[2]])));
+    prevKeydown = packet[latest];
+    Inline_Stack_Push(sv_2mortal(newSViv(usb_kbd_keycode[packet[latest]])));
     if(packet[0]) {
       packet[0] ^= packet[0] & 0xffffff00; // ugh
       // _dump_packet(packet);
@@ -135,21 +161,29 @@ void _keycode(SV* obj, int timeout) {
   }
   else {
     Inline_Stack_Push(sv_2mortal(newSViv(-1)));
+    if (latest <= 1) prevKeydown = 0;
   }
-	Inline_Stack_Done;		
+  hv_store((HV*)SvRV(obj), "prevKeydown", 11, newSViv((IV)prevKeydown), 0);
+  Inline_Stack_Done;
 }
 
 SV * _char(SV* obj) {
-  usb_dev_handle* handle = (usb_dev_handle*) SvIV(SvRV(obj));
+  int prevKeydown = fetchInt((HV*)SvRV(obj), "prevKeydown", 11, 0);
+  usb_dev_handle* handle = (usb_dev_handle*) fetchInt((HV*)SvRV(obj), "handle", 6, 0);
   SV * ans;
 
   char packet[PACKET_LEN];
-  int ret = usb_interrupt_read(handle, 0x81, packet, PACKET_LEN, 1000);
-  //fprintf(stderr, "read (%d)\n", ret);
-  // 0 is the shift code (maybe also something else)
-  // 2 is the scan code
-  if((ret > 0) && packet[2]) {
-    char c = code_to_key((packet[0] == 2), usb_kbd_keycode[packet[2]]);
+  int latest = usb_interrupt_read(handle, 0x81, packet, PACKET_LEN, 1000);
+
+  // find the most recent keydown (there could be several)
+  for (--latest; latest > 1 && packet[latest] == 0; --latest) {}
+
+  // if keydown found and is not a dupe (dupes happen with simultaneous keydowns)
+  if(latest > 1 && packet[latest] != prevKeydown) {
+    //fprintf(stderr, "read (%d)\n", ret);
+    // 0 is the shift code (maybe also something else)
+    prevKeydown = packet[latest];
+    char c = code_to_key((packet[0] == 2), usb_kbd_keycode[packet[latest]]);
     if(c) {
       const char str [2] = {c, '\0'};
       ans = newSVpvn( str, 1);
@@ -160,13 +194,14 @@ SV * _char(SV* obj) {
   }
   else {
     ans = newSVpvn( "", 0);
+    if (latest <= 1) prevKeydown = 0;
   }
-
+  hv_store((HV*)SvRV(obj), "prevKeydown", 11, newSViv((IV)prevKeydown), 0);
   return ans;
 }
 
 void _destroy(SV* obj) {
-  usb_dev_handle* handle = (usb_dev_handle*) SvIV(SvRV(obj));
+  usb_dev_handle* handle = (usb_dev_handle*) fetchInt((HV*)SvRV(obj), "handle", 6, 0);
   if(handle)
     cleanup(handle);
 }
